import os
import time
import json
import socket
import signal
import subprocess
from contextlib import contextmanager

__all__ = ['sagemaker_setup']

def sagemaker_setup():
    # Read info that SageMaker provides
    current_host = os.environ['SM_CURRENT_HOST']
    hosts = json.loads(os.environ['SM_HOSTS'])
    # Enable SSH connections between containers
    subprocess.Popen(["/usr/sbin/sshd", "-D"])
    if current_host == sorted(hosts)[0]:
        _wait_for_worker_nodes_to_start_sshd(hosts)
    else:
        sync_training_processes('dask-scheduler', current_host)

def _wait_for_worker_nodes_to_start_sshd(hosts, interval=1, timeout_in_seconds=360):
    with timeout(seconds=timeout_in_seconds):
        while hosts:
            print("hosts that aren't SSHable yet: %s", str(hosts))
            for host in hosts:
                ssh_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                if _can_connect(host, 22, ssh_socket):
                    hosts.remove(host)
            time.sleep(interval)

def _can_connect(host, port, s):
    try:
        print("testing connection to host %s", host)
        s.connect((host, port))
        s.close()
        print("can connect to host %s", host)
        return True
    except socket.error:
        print("can't connect to host %s", host)
        return False

class TimeoutError(Exception):
    pass

def sync_training_processes(proccess_id_string, worker_id, sync_frequency=300):
    training_process_started = False
    while True:
        time.sleep(sync_frequency)
        training_process_ps = subprocess.check_output(f'ps -elf | grep "{proccess_id_string}"', encoding='utf-8', shell=True)
        print(training_process_ps)
        training_process_count = subprocess.check_output(f'ps -elf | grep "{proccess_id_string}" | wc -l', encoding='utf-8', shell=True)
        training_process_count_str = training_process_count.replace("\n", "").strip()
        training_process_count = int(training_process_count_str) - 2
        training_process_running = training_process_count > 0
        if training_process_started:
            print(f'training processes running: {training_process_count}')
            if not training_process_running:
                print(f'Worker {worker_id} training completed.')
                time.sleep(5)
                return
        if not training_process_started:
            if training_process_running:
                training_process_started = True
            else:
                print(f'Worker {worker_id} exiting: training not started in 300 seconds.')
                return

@contextmanager
def timeout(seconds=0, minutes=0, hours=0):
    """
    Add a signal-based timeout to any block of code.
    If multiple time units are specified, they will be added together to determine time limit.
    Usage:
    with timeout(seconds=5):
        my_slow_function(...)
    Args:
        - seconds: The time limit, in seconds.
        - minutes: The time limit, in minutes.
        - hours: The time limit, in hours.
    """

    limit = seconds + 60 * minutes + 3600 * hours

    def handler(signum, frame):
        raise TimeoutError('timed out after {} seconds'.format(limit))

    try:
        signal.signal(signal.SIGALRM, handler)
        signal.setitimer(signal.ITIMER_REAL, limit)
        yield
    finally:
        signal.alarm(0)



import logging
from typing import Dict

import numpy as np
from sklearn.model_selection import ParameterGrid

from .local_searcher import LocalSearcher
from ..space import Categorical, Space, Int, Real

__all__ = ['LocalGridSearcher']

logger = logging.getLogger(__name__)


class LocalGridSearcher(LocalSearcher):
    """
    Grid Searcher that exhaustively tries all possible configurations. Grid with Int/Real ignores 'default' values in search spaces.
    """

    def __init__(self, grid_numeric_spaces_points_number=4, grid_num_sample_settings: Dict = None, **kwargs):
        """
        Parameters
        ----------
        grid_numeric_spaces_points_number: int, default = 4
            number of data point to sample from numeric space
        grid_num_sample_settings: dict (optional), default = None
            mapping between numeric space name and number of points to sample from it. Example {'a': 4} means
            sample 4 points from space 'a'. If no value present in this map, then `grid_numeric_spaces_points_number` will be used.
        """
        super().__init__(**kwargs)
        self._grid_numeric_spaces_points_number = grid_numeric_spaces_points_number
        self._grid_num_sample_settings = grid_num_sample_settings
        self._params_space = self._get_params_space()
        self._params_grid = ParameterGrid(self._params_space)
        self._grid_index = 0
        self._grid_length = len(self._params_grid)

    def _get_params_space(self) -> dict:
        param_space = dict()
        for key, val in self.search_space.items():
            if isinstance(val, Space):
                samples_num = self._get_samples_number(key)
                if isinstance(val, Int):
                    samples = min(val.upper - val.lower + 1, samples_num)
                    param_space[key] = np.linspace(val.lower, val.upper, samples, dtype=int)
                elif isinstance(val, Real):
                    space = np.geomspace if val.log else np.linspace
                    param_space[key] = space(val.lower, val.upper, num=samples_num)
                elif isinstance(val, Categorical):
                    sk = val.convert_to_sklearn()
                    param_space[key] = sk
                else:
                    raise AssertionError(f'Only Categorical is supported, but parameter "{key}" is type: {type(val)}')

        return param_space

    def _get_samples_number(self, key):
        samples = self._grid_numeric_spaces_points_number
        if self._grid_num_sample_settings is not None:
            samples = self._grid_num_sample_settings.get(key, samples)
        return samples

    def __len__(self):
        return self._grid_length - self._grid_index

    def get_config(self):
        """ Return new hyperparameter configuration to try next."""
        if len(self) <= 0:
            raise AssertionError(f'No configs left to get. All {self._grid_length} configs have been accessed already.')
        config = self._params_grid[self._grid_index]
        self._grid_index += 1
        for key, val in config.items():
            # If this isn't done, warnings are spammed in XGBoost
            if isinstance(val, np.int64):
                config[key] = int(val)
            elif isinstance(val, np.float64):
                config[key] = float(val)
        return config
